from rank_filter cimport lineRankOrderFilter1D_floating_inplace_loop

cimport cython

cimport numpy
import numpy

include "version.pxi"

numpy.import_array()


@cython.boundscheck(False)
def lineRankOrderFilter(numpy.ndarray a not None,
                        size_t half_length,
                        double rank,
                        int axis=-1,
                        numpy.ndarray out=None):
    """
        Runs a linear rank filter kernel along one dimension of an array.

        Args:
            a(numpy.ndarray):          array to run the rank filter over.

            half_length(size_t):       half the window size for the kernel.

            rank(double):              quantile to use from ``0.0`` to ``1.0``.

            axis(int):                 which axis to direct and run the kernel
                                       over defaults to the last one
                                       (e.g. ``-1``).

            out(numpy.ndarray):        where to store the results to. Creates a
                                       new array if not specified. If it is the
                                       same as a, it will run in-place.

        Returns:
            out(numpy.ndarray):        result of running the linear rank filter.
    """

    cdef int a_type_num = numpy.PyArray_TYPE(a)
    cdef int out_type_num

    cdef numpy.ndarray out_swap

    if -a.ndim <= axis < 0:
        axis += a.ndim
    elif not (0 <= axis < a.ndim):
        raise ValueError("`axis` needs to be within `a.ndim`")

    if not ((half_length + 1) <= <size_t>(a.shape[axis])):
        raise ValueError("Window must be no bigger than `a.shape[axis]`.")

    if not (0.0 <= rank <= 1.0):
        raise ValueError("The rank must be between 0.0 and 1.0.")

    if out is None:
        out = numpy.PyArray_NewCopy(a, numpy.NPY_CORDER)
        out_type_num = a_type_num
    else:
        out_type_num = numpy.PyArray_TYPE(out)
        if a_type_num != out_type_num:
            raise TypeError("Both `a` and `out` must have the same type.")
        if not numpy.PyArray_SAMESHAPE(a, out):
            raise ValueError("Both `a` and `out` must have the same shape.")
        if numpy.PyArray_CopyInto(out, a) == -1:
            raise RuntimeError("Unable to copy `a` to `out`.")

    out_swap = numpy.PyArray_SwapAxes(out, axis, out.ndim - 1)
    out_swap = numpy.PyArray_GETCONTIGUOUS(out_swap)

    cdef numpy.npy_intp out_swap_size = numpy.PyArray_SIZE(out_swap)
    cdef numpy.npy_intp out_swap_step = out_swap.shape[out.ndim - 1]
    if out_type_num == numpy.NPY_FLOAT32:
        lineRankOrderFilter1D_floating_inplace_loop[float](
            <float*>(out_swap.data), out_swap_size, out_swap_step,
            half_length, rank
        )
    elif out_type_num == numpy.NPY_FLOAT64:
        lineRankOrderFilter1D_floating_inplace_loop[double](
            <double*>(out_swap.data), out_swap_size, out_swap_step,
            half_length, rank
        )
    else:
        raise TypeError(
            "Only `float32` and `float64` are supported for `a` and `out`."
        )

    out_swap = numpy.PyArray_SwapAxes(out_swap, out.ndim - 1, axis)
    if numpy.PyArray_CopyInto(out, out_swap) == -1:
        raise RuntimeError("Unable to copy `out_swap` to `out`.")

    return(out)



cimport cython
from cython cimport floating

cimport numpy


cdef extern from "rank_filter.hxx" namespace "rank_filter":
    void lineRankOrderFilter1D[I1, I2](const I1& src_begin,
                                       const I1& src_end,
                                       I2& dest_begin,
                                       I2& dest_end,
                                       size_t half_length,
                                       double rank) nogil


@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.nonecheck(False)
cdef inline void lineRankOrderFilter1D_floating_inplace_loop(floating* out_data,
                                                             numpy.npy_intp out_size,
                                                             numpy.npy_intp out_step,
                                                             size_t half_length,
                                                             double rank) nogil:
    cdef numpy.npy_intp i

    cdef floating* out_begin = out_data
    cdef floating* out_end = out_data + out_step
    for i from 0 <= i < out_size by out_step:
        lineRankOrderFilter1D(
            out_begin, out_end, out_begin, out_end, half_length, rank
        )
        out_begin += out_step
        out_end += out_step



import os
import sys
from glob import glob

import setuptools
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext
from distutils.sysconfig import get_config_var, get_python_inc
from distutils.version import LooseVersion

import versioneer


assert LooseVersion(setuptools.__version__) >= LooseVersion("18.0"), \
        "Requires `setuptools` version 18.0 or higher."

class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

def readme():
    with open("README.rst", "r") as f:
        return(f.read())


version = versioneer.get_version()

with open("src/version.pxi", "w") as f:
     f.writelines([
         "__version__ = " + "\"" + str(version) + "\""
     ])

cython_dep = ["cython >= 0.23"]
numpy_dep = ["numpy >= 1.7"]
boost_dep = ["boost-cpp >= 1.56"]
boost_dep = (boost_dep if sys.argv[1] == "bdist_conda" else [])

setup_requires = cython_dep + numpy_dep
setup_requires = setup_requires if (sys.argv[1].startswith("bdist") or
                                    sys.argv[1].startswith("build") or
                                    sys.argv[1].startswith("install")) else []
build_requires = cython_dep + numpy_dep + boost_dep
install_requires = numpy_dep + boost_dep
install_requires += [] if sys.argv[1] == "bdist_conda" else cython_dep
tests_require = cython_dep + numpy_dep

include_dirs = [
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "include"),
    os.path.dirname(get_python_inc()),
    get_python_inc()
]
library_dirs = list(filter(
    lambda v: v is not None,
    [get_config_var("LIBDIR")]
))
sources = glob("src/*.pxd") + glob("src/*.pyx")

libraries = []
if os.name == "posix":
    libraries.append("boost_container")
elif os.name == "nt":
    libname = "boost_container"

    path = os.environ.get("LIB", "").split(";")

    libmatches = sum(
        list(glob(os.path.join(p, "%s*.lib" % libname)) for p in path), []
    )
    library_dirs.append(os.path.dirname(libmatches[0]))
    libraries.append(os.path.splitext(os.path.basename(libmatches[0]))[0])


extra_compile_args = []


setup(
    name="rank_filter",
    version=version,
    description="A simple python module containing an in-place linear rank"
                " filter optimized in C++.",
    long_description=readme(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: C++',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries'
    ],
    author="John Kirkham",
    author_email="kirkhamj@janelia.hhmi.org",
    url="https://github.com/nanshe-org/rank_filter",
    download_url="https://github.com/nanshe-org/rank_filter/archive/v%s.tar.gz"
                 % version,
    license="BSD",
    cmdclass=dict(
        list(versioneer.get_cmdclass().items()) +
        [
            ('build_ext', build_ext)
        ]
    ),
    setup_requires=setup_requires,
    build_requires=build_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite="tests",
    headers=glob("include/*.hxx"),
    ext_modules=[Extension("rank_filter",
                            sources=sources,
                            include_dirs=include_dirs,
                            library_dirs=library_dirs,
                            libraries=libraries,
                            extra_compile_args=extra_compile_args,
                            language="c++")],
    zip_safe=False
)
